﻿#region ファイル説明
//-----------------------------------------------------------------------------
// ModelImportDemoGame.cs
//=============================================================================
#endregion

#region Using ステートメント

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

#endregion

namespace ModelImportSample
{
    /// <summary>
    /// メタセコインポーターを使ったモデル表示デモ
    /// </summary>
    public class ModelImportSampleGame : Microsoft.Xna.Framework.Game
    {
        #region フィールド

        GraphicsDeviceManager graphics;

        // 表示するモデル
        Model model;

        // ローカル座標からワールド座標へ変換する行列リスト
        Matrix[] absoluteBoneTransforms;

        // 現在のモデル回転値
        float rotation = MathHelper.Pi;

        #endregion

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public ModelImportSampleGame()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

#if WINDOWS_PHONE
            graphics.PreferredBackBufferWidth = 480;
            graphics.PreferredBackBufferHeight = 800;

            // フルスクリーンにしてでWindows Phone 7上でバッテリー表示部分を消す
            graphics.IsFullScreen = true;

            // Windows Phone 7のGPUのリフレッシュレートは1/30秒なので
            // 更新頻度もそれに合わせる
            TargetElapsedTime = TimeSpan.FromTicks(333333);           
#else
            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;

            // アンチエリアスで綺麗に描画
            graphics.PreferMultiSampling = true;
#endif

        }

        /// <summary>
        /// コンテントの読み込み
        /// </summary>
        protected override void LoadContent()
        {
            // モデルの読み込み
            model = Content.Load<Model>("ino-oyako");

            // ローカル座標からワールド座標へ変換する行列を計算する
            absoluteBoneTransforms = new Matrix[model.Bones.Count];
            model.CopyAbsoluteBoneTransformsTo(absoluteBoneTransforms);

            // モデルにデフォルトライティングを適用
            foreach (ModelMesh mesh in model.Meshes)
                foreach (BasicEffect be in mesh.Effects)
                    be.EnableDefaultLighting();

            // テクスチャを使っている半透明MeshPartのエフェクトをAlphaTestEffectへ置き換える
            // メタセコイアインポーターは半透明描画が必要なMeshPart.TagにInt32値を設定するので
            // そのTagがnullかどうか調べることで半透明かどうかを判定できるようになっている。
            // 
            // この情報を使うことで不透明のMeshPartを描画した後に半透明のMeshPartを描画することで
            // 不透明、半透明メッシュ間での見た目の不具合を解決できる。
            // そのままでは半透明メッシュ間での見た目の不具合を解決できないので
            // ここではAlphaTestEffectに置き換えている。
            // 
            // ただし、AlphaTestEffectにはライティングを適用できないので、実際にはゲームの
            // 用途によって使い分ける必要がある

            var convertedEffects = new Dictionary<Effect, Effect>();
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                {
                    // このMeshPartが半透明でテクスチャを使っているか？
                    BasicEffect be = meshPart.Effect as BasicEffect;
                    if (meshPart.Tag == null || be.Texture == null)
                        continue;

                    // このエフェクトは処理済みか？
                    Effect effect;
                    if (!convertedEffects.TryGetValue(meshPart.Effect, out effect))
                    {
                        // AlphaTestEffectに置き換える
                        AlphaTestEffect ate = new AlphaTestEffect(GraphicsDevice);
                        ate.DiffuseColor = be.DiffuseColor;
                        ate.Texture = be.Texture;
                        ate.Alpha = be.Alpha;

                        // アルファ値が32以下のピクセルは破棄するようにする
                        ate.ReferenceAlpha = 32;

                        convertedEffects.Add(meshPart.Effect, ate);
                        effect = ate;
                    }

                    // 変換されたエフェクトを設定する
                    meshPart.Effect = effect;
                }
            }

        }

        /// <summary>
        /// ゲームの更新
        /// </summary>
        protected override void Update(GameTime gameTime)
        {
            // ゲーム終了処理
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // モデルの回転値を更新する
            float dt = (float)gameTime.ElapsedGameTime.TotalSeconds;
            rotation = MathHelper.WrapAngle(rotation + MathHelper.ToRadians(15) * dt);

            base.Update(gameTime);
        }

        /// <summary>
        /// ゲームの描画
        /// </summary>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(new Color(10, 9, 43));

            // プロジェクション、ビュー行列を生成する
            Matrix proj = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.ToRadians(30), GraphicsDevice.Viewport.AspectRatio, 1, 1000);

            Matrix rotMtx = Matrix.CreateRotationY(rotation);
            Vector3 camPos = Vector3.Transform(new Vector3(120, 50, 0), rotMtx);
            Matrix view = Matrix.CreateLookAt(camPos, new Vector3(0, 20, 0), Vector3.Up);

            // モデルの描画
            DrawModel(view, proj);

            base.Draw(gameTime);
        }

        #region モデル描画

        /// <summary>
        /// 指定されたMeshPartを描画する
        /// </summary>
        /// <param name="meshPart">描画するMeshPart</param>
        void DrawMeshPart(ModelMeshPart meshPart)
        {
            GraphicsDevice gd = meshPart.Effect.GraphicsDevice;

            meshPart.Effect.CurrentTechnique.Passes[0].Apply();

            gd.Indices = meshPart.IndexBuffer;
            gd.SetVertexBuffer(meshPart.VertexBuffer, meshPart.VertexOffset);

            gd.DrawIndexedPrimitives(PrimitiveType.TriangleList,
                0, 0, meshPart.NumVertices, meshPart.StartIndex,
                meshPart.PrimitiveCount);
        }

        /// <summary>
        /// モデルの描画
        /// </summary>
        /// <param name="view">ビュー行列</param>
        /// <param name="projection">プロジェクション行列</param>
        void DrawModel(Matrix view, Matrix projection)
        {
            GraphicsDevice gd = GraphicsDevice;

            // 不透明MeshPartを先に描画する
            GraphicsDevice.BlendState = BlendState.Opaque;
            foreach (ModelMesh mesh in model.Meshes)
            {
                // 行列を設定する
                foreach (IEffectMatrices effect in mesh.Effects)
                {
                    effect.World = absoluteBoneTransforms[mesh.ParentBone.Index];
                    effect.View = view;
                    effect.Projection = projection;
                }

                // 不透明なMeshPartを描画する
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                {
                    if (meshPart.Tag == null)
                        DrawMeshPart(meshPart);
                }
            }

            // 半透明MeshPartを描画する。
            // この時点で各エフェクトに必要な行列は既に設定されているので、再設定する必要はない
            gd.BlendState = BlendState.AlphaBlend;
            foreach (ModelMesh mesh in model.Meshes)
            {
                // 半透明MeshPartだけを描画する。
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                {
                    if (meshPart.Tag != null)
                        DrawMeshPart(meshPart);
                }
            }

        }

        #endregion

    }
}
